#!/bin/sh

#
# usage: ./packageinstall.sh package_identifier
# example: ./packageinstall.sh autoconf
#
# the second argument "user-only" can be given to perform a user-only installation
#
# Requirements:
# - a POSIX shell
# - wget or curl
#
# the script is assuming you're running it in the ports directory (which is a 
# bit stupid, need to be changed later) refer to "package definition 
# specifications.md" for laws and regulations
#
# NOTE: even if we assume no bugs in the packages installation scripts, there is a possibility for packages to be installed in a broken way, for example when the installation on the filesystem is interrupted half-way. The fact that installation of a package should be a transaction is not enforced by Axiom Ports. We decide not to address this problem, which, in practice is rare, and results only programs not working at all and in software not able to link at compile time, which is easy to see, does not result in security issues, does not impact other aleady installed software on the system
#
# TODO: authenticity is not done here (https alone does not guarantee authenticity because it's vulnerable to MITM attacks). What we should do is distribute a copy of the https certificate with Axiom Ports. That way, authenticity will be guaranteed, provided the user has an genuine copy of Axiom Ports.

set -ex   # -x because we want to have nice, explanatory, and verbose logs

packagedirectory="packages"   # the directory name where the packages definitions are located
installdirectory="/opt"   # default value for global installs, if the "user-only" argument is specified in $2, installdirectory becomes ~/.opt instead
bindirectory="/bin/"   # default value for global installs, if the "user-only" argument is specified in $2, becomes ~/.opt/bin/ instead
mirror_prefix="http://mirrors.tassig.com"   # a place to download tarballs of packages NOTE: we do not use an encrypted protocol here because privacy is not a concern by default, but integrity and authenticity on the other hand may be important and this is addressed by computing an optional cryptographic hash of the tarballs we download, the cryptographic hash being part of the package definition



# the function called by default to build a package
# if "custombuild()" is defined in the package definition, it will be used instead of this function
defaultbuild(){
	builddir="builddir-$package_fullname"  # build directory is "builddir" followed by the name of the package, this allows multiple builds of different software in parallel
	rm -rf $builddir
	mkdir -p $builddir   # do everything in builddir for tidiness
	cd $builddir
	
	if command -v wget &> /dev/null
	then
		wget -O archive $url   # TODO: use wget with --ca-certificate=$file, that way authenticity is guaranteed
	elif command -v curl &> /dev/null
	then
		curl -o archive $url
	else
		echo "we don't have wget or curl, we cannot proceed"
		return 2
	fi
	
	
	# verify integrity and authenticity of the archive if a sha256 hash has been provided
	if [ ! -z ${sha256+x} ] && [ `sha256sum archive | cut -d" " -f1` != "$sha256" ]
	then
		echo "The downloaded tarball does not match what we were supposed to download (the cryptographic hash does not match), and has likely been tampered. We abort the installation."
		exit 2
	fi


	tar xvf archive
	rm archive
	cd *   # cd into the package directory
	
	
	# we support cmake, meson, and autotools based builds
	# autotools has the priority over cmake or meson
	if [ -f CMakeLists.txt ] && [ ! -f configure ]
	then
		mkdir builddir
		cd builddir
		cmake -DCMAKE_INSTALL_PREFIX=$installdirectory/$package_fullname/ ..
	elif [ -f meson.build ] && [ ! -f configure ] 
	then
		meson .
	else
		./configure --prefix=$installdirectory/$package_fullname/
	fi 
	
	
	if [ -f "/proc/cpuinfo" ]
	then
		ncpu=`cat /proc/cpuinfo | grep processor | wc -l`   # on Linux
	else
		ncpu=`sysctl -n hw.ncpu`   # on many other POSIX systems
	fi
	
	
	make -j$ncpu
	if test -z $no_check   # run the make check, unless $no_check is set for this package definition
	then make -j$ncpu check || make -j$ncpu test
	fi
	make install
	
	
	ln -snv $package_fullname $installdirectory/$package_name || true   # if $package_name already exists (for example, another version is already installed), don't overwrite, and don't crash
	ln -sv $installdirectory/$package_name/bin/* $bindirectory || true   # don't crash if the links are already there, and don't overwrite these links
	if [ -d "$installdirectory/$package_fullname/lib/pkgconfig" ]
	then
		ln -svf $installdirectory/$package_fullname/lib/pkgconfig/* $installdirectory/pkgconf/lib/pkgconfig/  || true   # symlink pkg-config files, but don't crash if pkgconf is not installed, because for user installs it's not necessary
	fi
	if [ -d "$installdirectory/$package_fullname/share/pkgconfig" ]
	then   # some software also put pkg config files in share/ 
		ln -svf $installdirectory/$package_fullname/share/pkgconfig/* $installdirectory/pkgconf/lib/pkgconfig/
	fi
	
	# NOTE: we don't link .m4 files into autoconf, although we could. It's not necessary to build the vast majority of software so we don't do it.
	
	cd ../..
	rm -rf $builddir   # -f because for a normal user, a confirmation may be asked otherwise
}


# this does a bundle build (installation of the package and all its dependencies under one directory)
bundlebuild(){
	./bundlepackagebuild.sh $@
}



installpackage(){
	
	# the first argument is the package identifier
	package_full_id=$1
	
	# this argument ($package_full_id) can contain a version number, we are using the following heuristic to determine that:
	# - if there is a "-" in the package id
	# - and if after this "-", there is a digit as first character and there is a t list one "." in the last component
	last_component=$(echo $package_full_id | rev | cut -d"-" -f1 | rev)
	
	# check if the first character of $last_component is a digit and if there is a "." in $last_component, in which case we assume a version was provided, so we should install this specific version
	case $(echo $last_component | cut -c1) in 
		[0-9]* ) 
			case "$last_component" in
				*.*)
					provided_version=$last_component
					package_id=$(echo $package_full_id | rev | cut -d"-" -f2- | rev)
					;;
				*) 
					package_id=$package_full_id
					provided_version="0"
					;;
			esac
			;;
		* ) 
			package_id=$package_full_id
			provided_version="0"
			;;
	esac
	
	
	# sourcing package installer shell script, we have two types of packages:
	# - base packages in "base", where each package is a single file - shell script with name $1
	# - extended packages in "extended", where package is a shell script with name $1.sh
	# See "introduction to ports.md", which is the documentation
	if [ ! -f $packagedirectory/base/$package_id ]
	then
		. $packagedirectory/extended/$package_id.sh   # TODO: this file might not exist, exit with an error message, and do a search for packages with similar names
	else
		. $packagedirectory/base/$package_id   # source the file
	fi
	
	# if a version was provided, overwrite the default version of the package
	if [ "$provided_version" != "0" ]
	then
		echo "overriding version with $provided_version"
		package_version=$provided_version
		provided_version=0  # shall not be used anywhere else
	fi
	
	
	
	# if the package has been installed for the current user or globally, don't install again	
	if [ -d "$installdirectory/$package_name-$package_version" ] || [ -d "/opt/$package_name-$package_version" ] ||
	   [ -d "$installdirectory/$package_name-bundle" ]
	then
		echo "package $package_name is already installed, not reinstalling"
		return 0
	fi
	
	
	# if the package should be installed as a bundle, do so (but don't recurse again if we are already doing a bundle build, indicated by the use of --prefix)
	if [ "$bundle" = "yes" ] && [ "$2" != "--prefix" ]
	then
		echo "Doing a bundle build"
		bundlebuild $@
	fi
	
	
	echo "installing package $package_id"
	
	package_fullname=$package_name-$package_version 
	url=$mirror_prefix/$package_name/$package_fullname.tar.$tarball_suffix

	echo "build dependencies of $1: $build_dependencies"
	
	# installing the package's dependencies recursively
	for pkg_id in $build_dependencies
	do
		(./packageinstall.sh $pkg_id $2 $3) || exit $?   # You may want to use a recursive call to "installpackage" instead, but it's inheriting the shell local variables so spawning a new process allows us to start from something cleaner
	done
	
	# do a custom build if the package defines custombuild(), otherwise do a default build
	if type 'custombuild' 2>/dev/null | grep -q 'function'
	then custombuild
	else defaultbuild
	fi
	
	# call its postinstall() function if the package defines postinstall(), otherwise do nothing
	if type 'postinstall' 2>/dev/null | grep -q 'function'
	then postinstall
	fi
}




if [ "$2" = "user-only" ] || [ "$(id -u)" != "0" ]
then
	echo "Performing a user-only installation"
	installdirectory="$HOME/.opt"
	bindirectory="$HOME/.opt/bin/"
fi

if [ "$2" = "--prefix" ]
then
	newprefix=$3
	echo "Installation will be done in $newprefix"
	installdirectory=$3
fi


# bundle build vs normal build
if [ "$2" = "--bundle" ]
then
	echo "Doing a bundle build"
	bundlebuild $@
else
	installpackage $@
fi
