# How to debug flatpak applications

We distribute `flatpak` and `flatpak applications` via `hutchr` service. Flatpak installation is provided inside the `flatpak hutch`, a self contained flatpak distribution, built on an Axiom system. 

## How to install flatpak application

```
hutchr appID
```

where `appID` is flatpak application ID, refer to [https://docs.flatpak.org/en/latest/using-flatpak.html] under Identifiers. For example it may be `org.kde.kdenlive`


NOTE: `hutchr` follows this behavior:

  - if `flatpak hutch` was not installed it will be installed before `appID`
  - then `hutchr` makes call to flatpak hutch to install `appID`
  - `hutchr` performs addidional post installation steps (not important)
  - finally, `hutchr` run `appID` with appropriate call to flatpak


## What if application fails to run and has issues

Then we need to do debugging, by running `shell` inside flatpak application sandbox. Following steps are required:

  - run `shell` inside `flatpak hutch`
  - install matching SDK runtime for `appID`
  - run `shell` inside `appID` sandbox.


## How to run shell inside flatpak hutch

```
cd ~/.opt/hutches/flatpak  # this is where flatpak hutch gets installed
hutchr run /bin/sh
```

## How to install matching SDK runtime 

Flaptak applications get's installed with production `runtimes`, which has no `shell` and other useful tools. To be able to run `shell` we need to install matching runtime:

Example for `org.kde.kdenlive`:

```
flatpak run --command=sh --devel org.kde.kdenlive  # try to run shell

# at this point flatpak will report which SDK runtime is missing, so we install it

export XDG_DATA_DIRS=/opt/gsettings-desktop-schemas-3.20.0/share
flatpak install org.kde.Sdk/x86_64/5.13            # this installs required sdk runtime, 
```

## How to run shell inside application sandbox

```
export QT_QPA_PLATFORMTHEME=gtk3
flatpak run --command=sh --devel --device=all appID
```

NOTE: we use `QT_QPA_PLATFORMTHEME=gtk` workarouund for missing desktop portals, and `device=all` to expose complete `/dev` contents to the application we start.

In some cases we need to export `__GLX_VENDOR_LIBRARY_NAME=mesa` too, for example `us.zoom.Zoom`, to be able to manually star app, and use strace.

In case of debugging `us.zoom.Zoom`, to manually start application, we'll need to export one more variable `export LD_LIBRARY_PATH=/app/extra/zoom/`


Now we are in the `application sandxox` with SDK development runtime, which contains  useful tools for debugging, like `strace`, `gcc` and `gdb`. 

We will be able to re-compile, run and troubleshoot applications.

Each flatpak application comes with `manifest.json` file, where we can find all details about the application, git repository revisions, configure switches and dependences. This file is located inside `application sandbox` at: `/app/manifest.json`

Useful file is `.flatpak-info` in the root folder of `application sandbox`



# How to run shell with a specific runtime and version

In the moment we have more apps and runtimes installed, we are able to specify in which runtime we will run shell. This is useful when we want to get shell and do experiments in a specific runtime, regardless of the `appID`. It can be any `appID` which is already installed.

NOTE: These examples shows what hutchr does implicitly, when calling flatpak application

NOTE: some apps requires dbus daemon to run, start them with `dbus-launch`, the way hutchr does it, by calling `run.sh` inside flatpak hutch

Examples:

## run shell in application com.valvesoftware.Steam

```
# Explanations:
#  --command=sh - run shell in appllication sandbox
#  --filesystem=/home/me/ - share complete host user home folder
#  --device=all - share complete host /dev folder
#  --share=ipc - allow shared IPC between host and sandbox app, alsa dmix relies on it
#  --env=ALSA_CONFIG_DIR= - discard alsa envars  (some apps provide wrong path)
#  --env=ALSA_CONFIG_PATH=  - the same
#  --env=SDL_AUDIODRIVER=alsa - tell SDL2 library to use alsa as audio backend
#  --env=QT_QPA_PLATFORMTHEME=gtk3 - tell QT to use gtk3 dialog box

flatpak run --command=sh --filesystem=/home/me/ --device=all --share=ipc  --env=ALSA_CONFIG_DIR= --env=ALSA_CONFIG_PATH= --env=SDL_AUDIODRIVER=alsa --env=QT_QPA_PLATFORMTHEME=gtk3 com.valvesoftware.Steam
```

NOTE: without `--command=sh` flatpak will start application directly

NOTE: with `--devel` flag, flatpak will use `Sdk` runtime to start app, instead of `Platform`

## run shell in a given runtime and version

```
flatpak run --command=sh --filesystem=/home/me/ --device=all --share=ipc  --env=ALSA_CONFIG_DIR= --env=ALSA_CONFIG_PATH= --env=SDL_AUDIODRIVER=alsa --env=QT_QPA_PLATFORMTHEME=gtk3 org.freedesktop.Sdk//19.08
```


# How to get bwrap arguments

`flatpak` uses `bwrap` as backend to namespaces and cgropus. It is useful to print all arguments, which `flatpak` passes to `bwrap`. Use following command, to include bwrap args into more verbose output:

```
flatpak -vv run appId > out.debug

# out.debug will contain lot of debug prints, and we need to manually extract part with bwrap args, into bwrap.args file
# bwrap args are printed without newline, which is totally unreadable
# simple sed will make it usable
sed -i 's/--\+/\n&/g' bwrap.args
```


# How to completely remove flatpak hutch form the system

```
rm -rf ~/.opt/hutches/flatpak    # removes flatpak installation and runitems
rm -rf ~/.var/app                # removed user contets for installed flatpak applications

# manually remove all flatpak apps metadata in ~/.opt/hutches/
```


# How flatpak apps/runtimes adds more argumets, envars and extensions

Each faltpak application can define it's own:

  - specific command line argumens (like --devices=all, to share complete host /dev)
  - environment variables
  - flatpak runtime extensions, like `org.freedesktop.Platform.GL32` or `org.freedesktop.Platform.Compat.i386`

Application provider puts these requirements in application `metadata` file. In our case, in the flatpak hutch, it's located at `/opt/flatpak/var/lib/flatpak/app/appID/current/active/metadata`

These sections are of interest:

  - `[Context]` - place where run arguments are specified 
  - `[Environment]` - place where additional envars are specified
  - `[Extension *]` - runtime extensions


# How flatpak apps/runtimes, provides build and packaging info

Fpatpak app creators, for each app, provide details on application build process (repositories, commits, configs, etc...) and so called `modules`, other app dependencies, shipped with application itself.

All these information we can find in `manifest.json` file. In our case, in the flatpak hutch, it's located at `/opt/flatpak/var/lib/flatpak/app/appID/current/active/files/manifest.json`


# Flatpak support for 32-bit libs

Flatpak has 64-bit runtimes. Some of them has 32-bit extnsions (both Platform and Sdk). These 32-bit extension has `*.Compat.i386` in their filenames, for example:

  * `org.freedesktop.Platform.Compat.i386` (19.08)
  * `org.freedesktop.Sdk.Compat.i386`      (19.08)

List of runtimes that has 32-bit extension, can be obtained with this command:

`flatpak remote-ls | grep i386`

NOTE: Only following runtimes supports 32-bit libs: `org.freedesktop.Sdk//19.08`, `org.gnome.Sdk//3.34` and `org.gnome.Sdk//3.36`. Platform versions also. 

Extensions are not runrimes itself, we can't run commands like sh.  We can only run 64-bit runtime, with 32-bit extensions.


# How to compile and link 32-bit apps in the sandbox

To compile and link 32-bit apps, we need 32-bit toolchain. In case of freedesktop runtimes, we have to install it as a separate component:

`flatpak install org.freedesktop.Sdk.Extension.toolchain-i386`

NOTE: Gnome Compat.i386 extensions has 32-bit toolchain built in.

Now we need to run Sdk runtime (for which we have installed appropriate Sdk.Compat.i386 extension and toolchain), for example:

```sh
export SHAREDDIR=/home/jerry/shareddir # create it
export NAME=org.freedesktop.Sdk
export VERS=19.08

# NOTE: --allow=multiarch allows sandbox to run 32-bit applications (once we compile it)

flatpak run --command=sh --allow=multiarch --filesystem=$SHAREDDIR $NAME//$VERS
```

Inside sandbox, toolchain location is `/usr/lib/sdk/toolchain-i386/bin`.

Typically, to compile and link, we need to export these variables:

```sh
export PATH=$PATH:/usr/lib/sdk/toolchain-i386/bin
export CC=i386-unknown-linux-gnu-gcc
```


# NOTES:

Debugging flatpak applications is a complex process and may involve:

  - running diagnostic tools like `strace` and `gdb`
  - running flatpak on Debian based systems like Ubuntu, and comparing filesystem differences with Axiom system
  - patching and recompiling `flatpak` sources
  - writing and compiling test programs
  - manually editing application sandbox contents


More info at: [https://docs.flatpak.org/en/latest/debugging.html]
