# How to migrate Axiom pathes

With new versions of flatpak, we will have to migrate axiom patches. Most likely, musl patch will remain the same.

For each patch, we point to more detailed documentation.

This is the procedure:

  1. first step is to make new version of flatpak to build without any issues, with musl patch. For this step we have to modify "flatpak.sh"
  2. Inject souce code snippets, into flatpak sources, following provided instructions.
  3. Compile and test patched flatpak
  4. export patches 

## Make copies of original source files

```sh
cp common/flatpak-utils.c common/flatpak-utils.c.orig
cp app/flatpak-cli-transaction.c app/flatpak-cli-transaction.c.orig
cp common/flatpak-run.c common/flatpak-run.c.orig
```

## xdg-runtime-dir-fix.patch

For more details, refer to `documentation/flatpak issues.md`.

In the file `common/flatpak-utils.c`, we patch function `char * flatpak_get_real_xdg_runtime_dir (void)`. Origially it is:

```c
char *
flatpak_get_real_xdg_runtime_dir (void)
{
  return realpath (user_runtime_dir, NULL);
}
```

Replace it with:

```
char *
flatpak_get_real_xdg_runtime_dir (void)
{
  /* Axiom patch: xdg-data-dir */
  /* Originally this function was returning "realpath (g_get_user_runtime_dir (), NULL)"
   *
   * According to glib specification, g_get_user_runtime_dir() will return XDG_RUNTIME_DIR
   * If XDG_RUNTIME_DIR was not set, it will return g_get_user_cache_dir()
   *
   * But in case XDG_RUNTIME_DIR is set but not a valid path, NULL will be returned,
   * empty path will be passed to bwrap arguments, and flatpak will crash. The same happens
   * if g_get_user_cache_dir() is not a valid path.
   *
   * Simplest workaround is to add additional check, and make sure we create user runtime folder
   *
   */
  const char *user_runtime_dir = g_get_user_runtime_dir();

  if (realpath (user_runtime_dir, NULL) == NULL)
  {
    if (g_mkdir_with_parents (user_runtime_dir, 0755))
        g_debug ("Can't create user runtime dir: %s\n", user_runtime_dir);
  }
  /* end Axiom patch*/

  return realpath (user_runtime_dir, NULL);
}
```

## mainscreen-progress.patch

For more details, refer to `documentation/mainscreen progress.md`.

Open file `app/flatpak-cli-transaction.c`

### hunk 1

We inject set of additional top level functions. Following code shall be inserted below `format_duration` function:

```c
/* Axiom patch: report installation progress to mainscreen */
/*
 * Application installation is called 'transaction', and has several steps called `operations`
 * It is because flatpak can install various application runtime dependencies.
 * To report total progress we apply a patch which do following:
 * 
 *   - opens mainscreen socket
 *   - calculate num_of_operations - total number of operations we have to do
 *   - use `progress_changed_cb` signal handler to read percentage progress of an operation
 *   - calculate total percentage and write to mainscreen socket
 * 
 * IMPORTANT NOTE:
 * This is less preferred method then calculating bytes, but there is an issue with such
 * approach. We could calculate total transaction size in bytes, and use transfered
 * bytes to calculate percentage. the issue with this approach is that some operations
 * reports significantly bigger download size that it will be actually downloaded.
 * These are runtime dependences with .Local extension, denoted as `partial`. 
 */

int num_of_operations; 

int open_mainscreen_socket (void);
int serialize_int_4 (char *buff, int val);
int serialize_string (char *buff, char *string);
void write_mainsreen_socket_progress (int progress);
void write_mainsreen_socket_buff (char * wr);


#include <sys/socket.h> /* socket () */
#include <sys/un.h>     /* struct sockaddr_un */

#define MAINSCREEN_SOCKET_PATH "/tmp/mainscreen-ipc.unix"
#define MAINSCREEN_PROGRESS_MESSAGE_TYPE 1

int fd;           /* socket fd */
char wrbuf[512];  /* buffer used to compose message for the mainscreen */
char app_id[512]; /* flatpak application ID */

int open_mainscreen_socket (void)
{
    struct sockaddr_un addr;
    
    fd = 0;
    if((fd = socket (AF_UNIX, SOCK_DGRAM, 0)) == -1)
        return -1;

    /* setup unix sokcket file path */
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    strcpy(addr.sun_path, MAINSCREEN_SOCKET_PATH);

	return connect(fd, (struct sockaddr*)&addr, sizeof(addr));
}

/* pack integer of 4 bytes into a buffer, little endian order */
int serialize_int_4 (char *buff, int val)
{
    int i, pos=0;
    for (i=3; i>=0; i--)
        buff[pos++] = (0xff << (i*8)) & val;
    return 4;
}

/* pack string into a buffer */
int serialize_string (char *buff, char *string)
{
    int i, len=strlen(string);
    for (i=0; i < len; i++)
        buff[i]=string[i];
    return len;
}

void write_mainsreen_socket_progress (int progress)
{
    wrbuf[0] = MAINSCREEN_PROGRESS_MESSAGE_TYPE;
    int pos = 1;
    pos += serialize_int_4 (wrbuf + pos, strlen(app_id));
    pos += serialize_string (wrbuf + pos, app_id);
    pos += serialize_int_4 (wrbuf + pos, progress);

    if (fd)
        write (fd, wrbuf, pos);
}

/* deobug function */
void write_mainsreen_socket_buff (char * wr)
{
    if (fd)
        write (fd, wrbuf, strlen (wr));
}
/* end Axiom patch */
```

### hunk 2

Patch in the function `progress_changed_cb`.

Find this code part:

```c
  guint64 max = flatpak_transaction_operation_get_download_size (op);

  if (elapsed_time > 0)
```

Insert code between these lines:

```c
  /* Axiom patch: total progress report to mainscreen */
  /*
   * Operations are done one by one, sequentially
   * This handler get's called each time progress of an operation changes
   * We can get current operation in `current_op` (an integer which starts form 0)
   * We already know how many of them will be, num_of_operations
   * 
   * We have followind data available:
   *   `num_of_operations`, will have N*100 percent to be done.
   *   `current_operation` [0..N-1]
   *   `percent` is current operation progress [0-100]
   * 
   * total_progress_percent = ((current_operation*100) + percent)/num_of_operations
   */

  /* not likely we can get division by zero, but double check */
  if (num_of_operations)
  {
    int current_operation = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (op), "row"));
    int total_progress_percent =((current_operation*100) + percent)/num_of_operations;
    
    write_mainsreen_socket_progress (total_progress_percent);
    
    /* debug code
    sprintf(wrbuf, "[%d] [%d] [%d] [%d]",
        total_progress_percent,
        num_of_operations,
        current_operation, 
        percent);
    write_mainsreen_socket_buff ( wrbuf );
    */
  }
  /* end Axiom patch */
```

### hunk 3

Patch in the function `transaction_ready`

Find this part:

```c
  for (l = ops, i = 1; l != NULL; l = l->next, i++)
    {
      FlatpakTransactionOperation *op = l->data;
      FlatpakTransactionOperationType type = flatpak_transaction_operation_get_operation_type (op);
      const char *ref = flatpak_transaction_operation_get_ref (op);
      const char *remote = flatpak_transaction_operation_get_remote (op);
      g_auto(GStrv) parts = flatpak_decompose_ref (ref, NULL);
      g_autofree char *rownum = g_strdup_printf ("%2d.", i);

      flatpak_table_printer_add_column (printer, rownum);
      flatpak_table_printer_add_column (printer, "   ");
      flatpak_table_printer_add_column (printer, parts[1]);
      flatpak_table_printer_add_column (printer, parts[2]);
      flatpak_table_printer_add_column (printer, parts[3]);

      if (self->installing + self->updating + self->uninstalling > 1)
        flatpak_table_printer_add_column (printer, op_shorthand[type]);

      if (type == FLATPAK_TRANSACTION_OPERATION_INSTALL ||
          type == FLATPAK_TRANSACTION_OPERATION_INSTALL_BUNDLE ||
          type == FLATPAK_TRANSACTION_OPERATION_UPDATE)
        {
          g_autoptr(FlatpakRef) rref = flatpak_ref_parse (ref, NULL);
          guint64 download_size;
          g_autofree char *formatted = NULL;
          g_autofree char *text = NULL;

          download_size = flatpak_transaction_operation_get_download_size (op);
          formatted = g_format_size (download_size);

          flatpak_table_printer_add_column (printer, remote);
          if (g_str_has_suffix (flatpak_ref_get_name (rref), ".Locale"))
            text = g_strdup_printf ("< %s (%s)", formatted, _("partial"));
          else
            text = g_strdup_printf ("< %s", formatted);
          flatpak_table_printer_add_decimal_column (printer, text);
        }

      g_object_set_data (G_OBJECT (op), "row", GINT_TO_POINTER (flatpak_table_printer_get_current_row (printer)));
      flatpak_table_printer_finish_row (printer);
    }
```

Insert code below:

```c
  /* Axiom patch: progress report to mainscreen */
  num_of_operations = i - 1;

  /* get flapak application ID, we need to send it compose mainscreen progress message */
  for (l = ops; l != NULL; l = l->next)
    {
      FlatpakTransactionOperation *op = l->data;
      const char *ref = flatpak_transaction_operation_get_ref (op);
      g_autoptr(FlatpakRef) rref = flatpak_ref_parse (ref, NULL);

      if (flatpak_ref_get_kind (rref) == FLATPAK_REF_KIND_APP)
        g_stpcpy (app_id, flatpak_ref_get_name (rref));
    }
  g_print ("app_id=%s\n", app_id);
  if (open_mainscreen_socket())
    g_print ("mainscreen socket status: open fail!\n");
  else
    g_print ("mainscreen socket status: open success!\n");
  /* end Axiom patch */
```


## runtime-and-application-fixes.patch

For more details, refer to `documentation/application and runtime fixes.md`.

Open file `common/flatpak-run.c`

### hunk 1

We inject set of additional top level functions. Following code shall be inserted below `flatpak_run_add_pulseaudio_args` function:

```c
/* Axiom patch: runtime fixes */

/*
 * NOTE: we use rf_ prefix for variables related to runtime fixes */

/* Global variable explanation:
 *
 *   rf_app_ref - full application reference in format TYPE/NAME/ARCH/VERSION
 *   rf_runtime_ref - full runtime reference in format TYPE/NAME/ARCH/VERSION
 *   rf_app_metakey - pointer to application metafile, used to read data form it
 *   rf_runtime_metakey - pointer to runtime metafile, used to read data form it
 *
 * These variables will be filled in flatpak_run_app
 */
gchar *rf_app_ref = NULL;
gchar *rf_runtime_ref = NULL;
GKeyFile *rf_app_metakey = NULL;
GKeyFile *rf_runtime_metakey = NULL;


/* auxilliary function, to find files on a given PATH, which contains matching PATTERN
 * returns a list of found files, or NULL
 */
static GList *
find_files_with_matching_pattern (const char *path, char *pattern)
{
    GList *res = NULL;
   	DIR *dir = NULL;
    struct dirent *dirent = NULL;

    dir = opendir(path);
    if (!dir)
        return NULL;

    while ((dirent = readdir( dir)))
    {
        if (strstr(dirent->d_name, pattern))
            res = g_list_append (res, g_strdup(dirent->d_name));
    }
    return res;
}

/* auxilliary function, to recursively find all regular files on a given PATH
 * returns a list of found files, or NULL
 */
static void
find_rec_all_files (const char *path, GList **fixes)
{
   	DIR *dir = NULL;
    struct dirent *dirent = NULL;
    gchar *rec_path;

    dir = opendir(path);
    if (!dir)
        return;

    while ((dirent = readdir( dir)))
    {
        if (strcmp(dirent->d_name, ".") != 0 && strcmp(dirent->d_name, "..") != 0)
        {
          if (dirent->d_type == DT_DIR)
          {
            rec_path = g_strdup_printf ("%s/%s", path, dirent->d_name);
            find_rec_all_files (rec_path, fixes);
            if (rec_path)
              g_free(rec_path);
          }
          
          if (dirent->d_type == DT_REG)
            *fixes = g_list_append (*fixes, g_strdup_printf ("%s/%s", path, dirent->d_name));
        }
    }
    return;
}

/* auxilliary function to find and read runtime library symlink, based on library full name
 * returns libname pointed by a symlink, or NULL
 *
 * we use it when flatpak decides to update their runtime libraries
 * appropriate runtime symlink will point to the new library version
 * so we use old library name, to find "soname", which is a symlink, and read it's value
 *
 * Example: we search based on this name libSDL2-2.0.so.0.10.0
 *          soname=libSDL2-2.0.so.0
 *          check if soname symlink exist, and read it
 *          for example, it may be libSDL2-2.0.so.0 -> libSDL2-2.0.so.0.12.0
 *          so we return libSDL2-2.0.so.0.12.0
 */
static char *
find_updated_library_name_form_current (char *path, char *lib_name_full)
{
    /* get soname by copying and cutting lib_name_full */
    gchar *soname = g_strdup (lib_name_full);
    char *ptr = strstr(soname, ".so.");
    ptr += 4;
    for (int i=0; i < 3; i++) /* we don't really expect library version with 4 digits */
      if ( *(ptr+i) == '.' )
        *(ptr+i) = 0;

    gchar *soname_full_path = g_strdup_printf ("%s/%s", path, soname);
    g_free(soname);

    struct stat sb;
    if ((lstat(soname_full_path, &sb) == 0 && S_ISLNK(sb.st_mode)))
    {
        char linkval [PATH_MAX];
        size_t len = readlink (soname_full_path, linkval, sizeof(linkval) - 1);
        if (len != -1)
        {
            linkval[len]=0; /* !! readlink don't provide 0 termination */
            return g_strdup(linkval);
        }
    }

    g_free (soname_full_path);
    return NULL;
}

/* auxilliary function to check file existance
 * before bind mounting our fix into ro mounted runtimes, we need to confirm it
 * exists in flatpak runtime
 *
 * return value:
 *   0 - if file exist, and is regular file
 *  -1 - if file does not exist
 */
static int
check_file_existance(char *path, char *dbg)
{
    struct stat sb;
    if ((stat(path, &sb) == 0 && S_ISREG(sb.st_mode)))
      return 0;

    g_debug ("%s: FILE NOT FOUND: %s", dbg, path);
    return -1;
}


/* auxilliary function to check folder existance
 * before bind mounting our fix into ro mounted runtimes, we need to confirm it
 * exists in flatpak runtime
 *
 * return value:
 *   0 - if folder exist
 *  -1 - if folder does not exist
 */
static int
check_folder_existance(char *path, char *dbg)
{
    struct stat sb;
    if ((stat(path, &sb) == 0 && S_ISDIR(sb.st_mode)))
      return 0;

    g_debug ("%s: FOLDER NOT FOUND: %s", dbg, path);
    return -1;
}

/* auxilliary function, bind mounts all matching ".so." files in a given LIB_PATH,
 * to DEST folder in the sandbox application
 *
 * Explanations:
 *
 *   char *lib_path - path in "runtime fixes" where fixed libs are located
 *   char *flatpak_runtime_install_path - path in flatpak installed runtimes where original libs are locates, we need to match fixed library names with original
 *   char *dest - path in the flatpak sandbox, where fixed libs will be mounted
 *   char *dbg - string to identify who call this function, in logs
 */
static void
mount_all_lib_files (FlatpakBwrap *bwrap, char *lib_path, char *flatpak_runtime_install_path, char *dest, char *dbg)
{
    GList *libs = find_files_with_matching_pattern(lib_path, ".so.");
    GList *l;

    if (libs)
    {
        for (l = libs; l != NULL; l = l->next)
        {
            gchar *lib_name = l->data;
            gchar *source = g_strdup_printf ( "%s/%s",
                                lib_path,
                                lib_name);

            /* IMPORTANT: for each lib we are trying to bind mount, we have to check
             *            if it exists in the flatpak runtime installation path.
             *
             *  flatpak runtimes get's async updates form time to time, anb
             *  it may happen that we provide fixed lib version N, and new
             *  update brings newer lib version.
             *
             *  In such case we apply strategy, to bind mount our older version of lib
             *  over the newer one, and we also need to update runtime symlink. We rely
             *  on the fact that libs are keeping backward compatibility
             *
             *  Example: org.freedesktop.Platform/19.08 once provided SDL2 lib:
             *    libSDL2-2.0.so.0 -> libSDL2-2.0.so.0.10.0
             *
             *  for which we provide a fix libSDL2-2.0.so.0.10.0, in runtime fixes.
             *
             *  At one point this runtime get's an update, and now we have:
             *    libSDL2-2.0.so.0 -> libSDL2-2.0.so.0.12.0
             *
             *  To allow apps work properly, we need to detect updated lib presence
             *  and bind mount our libSDL2-2.0.so.0.10.0 over runtime libSDL2-2.0.so.0.12.0
             *
             *  IMPORTANT NOTE: we rely on same soname, to belive backward compltibility
             *                  was not broken, in this case libSDL2-2.0.so.0
             */
            gchar *target = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, lib_name);

            if (check_file_existance(target, dbg) == 0)
            {
                /* this is 99% percent of cases, we have found matching file
                 * in the specified flatpak runtime, so it's safe to bind mount
                 * our fixed library, with bwrap call
                 */
                gchar *destfile = g_strdup_printf ("%s/%s", dest, lib_name);
                g_debug("%s: mounting [%s] to [%s]", dbg, source, destfile);

                flatpak_bwrap_add_args (bwrap,
                                    "--ro-bind",
                                    source,
                                    destfile,
                                    NULL);
                g_free (destfile);
            }
            else
            {
                /* now this is the case when we can't find matching file in the
                 * specified flatpak runtime, this is most likely because flatpak
                 * decided to update this library.
                 *
                 * we use library name, whih we provide in runtime fixes and call
                 * a function which will try to find new library name, in the specified
                 * runtime.  */
                gchar *updated_target = find_updated_library_name_form_current (flatpak_runtime_install_path, lib_name);

                if (updated_target)
                {
                    /* we have found newer library than we provide fix to */
                    /* NOTE: find_updated_library_name_form_current will also test for
                     *       updated file existance, so it's safe to bind mount */
                    g_debug("%s: FOUND UPDATED [%s]", dbg, updated_target);

                    gchar *destfile = g_strdup_printf ("%s/%s", dest, updated_target);
                    g_debug("%s: mounting [%s] to [%s]", dbg, source, destfile);

                    flatpak_bwrap_add_args (bwrap,
                                    "--ro-bind",
                                    source,
                                    destfile,
                                    NULL);
                    g_free (destfile);
                    g_free (updated_target);
                }
            }
            g_free (lib_name);
            g_free (source);
            g_free (target);
        }
        g_list_free (libs);
    }
}


/* ALSA FIXES: */
/* The reason we fix alsa-lib is because runtime provided alsa-lib has no support of any
 * useful plugin like: dmix, async, softvol, etc... So we provide our own re-compiled version
 *
 * Explanations:
 *  gchar *runtime_ref - runtime reference, in format "type/name/arch/version"
 *  gchar *flatpak_runtime_install_path - path where specified runtime is installed on the host, so we can check library/other file existance
 *  gchar *mount_point_x86_64 - path in the app sandbox where 64bit libs fixes will be bind mounted
 *  gchar *mount_point_i386 - the same for 32bit libs
 */
static void
flatpak_run_add_alsa_support_args (FlatpakBwrap *bwrap, gchar *runtime_ref, gchar *flatpak_runtime_install_path, gchar *mount_point_x86_64, gchar *mount_point_i386)
{
    /* split runtime_ref string into parts:
     *   runtime_parts[0] - type { app, runtime }, here it should be always runtime
     *   runtime_parts[1] - runtime name
     *   runtime_parts[2] - runtime arch
     *   runtime_parts[3] - runtime version
     */
    g_auto(GStrv) runtime_parts = g_strsplit (runtime_ref, "/", 0);

    gchar *source = NULL;  /* runtime fixes paths, where fixes are stored */
    gchar *target = NULL;  /* flatpak runtie paths, contents that we try to fix */

    /* some runtimes provides 32bit libs, as runtime extension
     * and if these extensions are not installed in flatpak
     * there is no point in mounting fixes */
    if (mount_point_i386)
    {
        gchar *full_32_runtime_fix_path = g_strdup_printf ("/opt/flatpak/runtime fixes/%s/%s/i686/alsa", runtime_parts[1], runtime_parts[3]);

        g_debug("ALSA FIX: 32-bit full_32_runtime_fix_path = %s", full_32_runtime_fix_path);

        /* check if we provide 32-bit fix for given runtime */
        if (check_folder_existance (full_32_runtime_fix_path, "ALSA FIX") == 0)
        {
            /* override runtime library (or libraries), with our re-compiled versions
             * NOTE: we will find library full name by looking inti "runtime fixes"
             *       and we'll mount what we find. In general, different runtimes has
             *       different version of libs
             */
            gchar *lib_32_path = g_strdup_printf ("%s/lib", full_32_runtime_fix_path);

            /* use aux function to find and mount all library files */
            mount_all_lib_files (bwrap,
                                 lib_32_path,
                                 flatpak_runtime_install_path,
                                 mount_point_i386,
                                 "ALSA FIX");

            g_free (lib_32_path);
            g_free (full_32_runtime_fix_path);
        }
        else
            g_debug ("ALSA FIX: 32-bit [%s] not supported", full_32_runtime_fix_path);
    }


    if (mount_point_x86_64)
    {
        gchar *full_64_runtime_fix_path = g_strdup_printf ("/opt/flatpak/runtime fixes/%s/%s/%s/alsa", runtime_parts[1], runtime_parts[3], runtime_parts[2]);

        g_debug("ALSA FIX: full_64_runtime_fix_path = %s", full_64_runtime_fix_path);

        /* check if we provide alsa fix for given runtime */
        if (check_folder_existance (full_64_runtime_fix_path, "ALSA FIX") == 0)
        {
            target = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, "etc/alsa");

            if (check_folder_existance (target, "ALSA FIX") == 0)
            {
               /* override runtime /etc/alsa, with an empty folder */
                source = g_strdup_printf ("%s/etc/alsa", full_64_runtime_fix_path);
                g_debug("ALSA FIX: mounting [%s] to [/etc/alsa]", source);
                flatpak_bwrap_add_args (bwrap,
                                  "--ro-bind", source, "/etc/alsa",
                                  NULL);
                g_free (source);
            }
            g_free (target);

            /* override runtime /usr/share/alsa, with contents form fixed recompiled alsa-lib */
            target = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, "share/alsa");

            if (check_folder_existance (target, "ALSA FIX") == 0)
            {
                source = g_strdup_printf ("%s/share/alsa", full_64_runtime_fix_path);
                g_debug("ALSA FIX: mounting [%s] to [/usr/share/alsa]", source);
                flatpak_bwrap_add_args (bwrap,
                                      "--ro-bind", source , "/usr/share/alsa",
                                      NULL);
                g_free (source);
            }
            g_free (target);

            /* mount host alsa config, where runtime alsa-lib will expect it */
            target = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, "share/alsa/alsa.conf.d");

            if (check_folder_existance (target, "ALSA FIX") == 0)
            {
              g_debug("ALSA FIX: mounting [/opt/alsa-lib/share/alsa/alsa.conf.d] to [/usr/share/alsa/alsa.conf.d]");
              flatpak_bwrap_add_args (bwrap,
                                    "--ro-bind", "/opt/alsa-lib/share/alsa/alsa.conf.d", "/usr/share/alsa/alsa.conf.d",
                                    NULL);
            }
            g_free (target);

            /* override runtime library (or libraries), with our re-compiled versions
             * NOTE: we will find library full name by looking inti "runtime fixes"
             *       and we'll mount what we find. In general, different runtimes has
             *       different version of libs
             */
            gchar *lib_64_path = g_strdup_printf ("%s/lib", full_64_runtime_fix_path);
            gchar *flatpak_runtime_install_path_lib = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, &mount_point_x86_64[5]);

            /* use aux function to find and mount all library files */
            mount_all_lib_files (bwrap,
                                 lib_64_path,
                                 flatpak_runtime_install_path_lib,
                                 mount_point_x86_64,
                                "ALSA FIX");
            g_free (lib_64_path);

            /* NOTE: This part is for troubleshooting only*/
            /* in case of devel runtime (--devel flag) or explicit Sdk runtime, override
             * includes, and *.pc file, to match replaced lib, so code examples can be compiled.
             *
             * NOTE: includes and .pc files are same for 32bit and 64bit libs
             */
            if (g_strrstr (full_64_runtime_fix_path, "Sdk"))
            {
                source = g_strdup_printf ("%s/include/alsa", full_64_runtime_fix_path);
                g_debug("ALSA FIX: mounting [%s] to [/usr/include/alsa]", source);
                flatpak_bwrap_add_args (bwrap,
                                  "--ro-bind", source, "/usr/include/alsa",
                                  NULL);
                g_free (source);

                /* get the path where alsa.pc is located, it differs for certain runtimes */
                source = g_strdup_printf ("%s/lib/pkgconfig/alsa.pc", full_64_runtime_fix_path);

                gchar pc_path[256];
                sprintf(pc_path, "%s/pkgconfig/alsa.pc", mount_point_x86_64);
                g_debug("ALSA FIX: mounting [%s] to [%s]", source, pc_path);

                flatpak_bwrap_add_args (bwrap,
                                  "--ro-bind", source, pc_path,
                                  NULL);
                g_free (source);
            }
            g_free (full_64_runtime_fix_path);
        }
        else
            g_debug ("ALSA FIX: %s not supported", full_64_runtime_fix_path);
    }
}



/* Runtime provided sdl2 lib was compiled without alsa support, so we provide our re-compiled version
 */
static void
flatpak_run_add_sdl2_support_args (FlatpakBwrap *bwrap, gchar *runtime_ref, gchar *flatpak_runtime_install_path, gchar *mount_point_x86_64, gchar *mount_point_i386)
{
    /* split runtime_ref string into parts:
     *   runtime_parts[0] - type { app, runtime }, here it should be always runtime
     *   runtime_parts[1] - runtime name
     *   runtime_parts[2] - runtime arch
     *   runtime_parts[3] - runtime version
     */
    g_auto(GStrv) runtime_parts = g_strsplit (runtime_ref, "/", 0);

    /* some runtimes provides 32bit libs, as runtime extension
     * and if these extensions are not installed in flatpak
     * there is no point in mounting fixes */
    if (mount_point_i386)
    {
        gchar *full_32_runtime_fix_path = g_strdup_printf ("/opt/flatpak/runtime fixes/%s/%s/i686/sdl2", runtime_parts[1], runtime_parts[3]);

        g_debug("SDL2 FIX: 32-bit full_32_runtime_fix_path = %s", full_32_runtime_fix_path);

        /* check if we provide 32-bit fix for given runtime */
        if (check_folder_existance (full_32_runtime_fix_path, "SDL2 FIX") == 0)
        {
            /* override runtime library (or libraries), with our re-compiled versions
             * NOTE: we will find library full name by looking inti "runtime fixes"
             *       and we'll mount what we find. In general, different runtimes has
             *       different version of libs
             */
            gchar *lib_32_path = g_strdup_printf ("%s/lib", full_32_runtime_fix_path);

            /* use aux function to find and mount all library files */
            mount_all_lib_files (bwrap,
                                 lib_32_path,
                                 flatpak_runtime_install_path,
                                 mount_point_i386,
                                "SDL2 FIX");

            g_free (lib_32_path);
            g_free (full_32_runtime_fix_path);
        }
        else
            g_debug ("SDL2 FIX: 32-bit [%s] not supported", full_32_runtime_fix_path);
    }

    if (mount_point_x86_64)
    {
        gchar *full_64_runtime_fix_path = g_strdup_printf ("/opt/flatpak/runtime fixes/%s/%s/%s/sdl2", runtime_parts[1], runtime_parts[3], runtime_parts[2]);

        g_debug("SDL2 FIX: full_64_runtime_fix_path = %s", full_64_runtime_fix_path);

        /* check if we provide 64-bit fix for given runtime */
        if (check_folder_existance (full_64_runtime_fix_path, "SDL2 FIX") == 0)
        {
            /* override runtime library (or libraries), with our re-compiled versions
             * NOTE: we will find library full name by looking inti "runtime fixes"
             *       and we'll mount what we find. In general, different runtimes has
             *       different version of libs
             */
            gchar *lib_64_path = g_strdup_printf ("%s/lib", full_64_runtime_fix_path);
            gchar *flatpak_runtime_install_path_lib = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, &mount_point_x86_64[5]);

            /* use aux function to find and mount all library files */
            mount_all_lib_files (bwrap,
                                 lib_64_path,
                                 flatpak_runtime_install_path_lib,
                                 mount_point_x86_64,
                                 "SDL2 FIX");
            g_free (lib_64_path);

            /* NOTE: This part is for troubleshooting only*/
            /* in case of devel runtime (--devel flag) or explicit Sdk runtime, override
             * includes, and *.pc file, to match replaced lib, so code examples can be compiled.
             *
             * NOTE: includes and .pc files are same for 32bit and 64bit libs
             */
            if (g_strrstr (full_64_runtime_fix_path, "Sdk"))
            {
                gchar *source = g_strdup_printf ("%s/include/SDL2", full_64_runtime_fix_path);

                g_debug("SDL2 FIX: mounting [%s] to [/usr/include/SDL2]", source);

                flatpak_bwrap_add_args (bwrap,
                                        "--ro-bind",
                                        source,
                                        "/usr/include/SDL2",
                                        NULL);
                g_free (source);

                /* get the path where sdl2.pc is located, it differs for certain runtimes */
                source = g_strdup_printf ("%s/lib/pkgconfig/sdl2.pc", full_64_runtime_fix_path);

                gchar *pc_path = g_strdup_printf ("%s/pkgconfig/sdl2.pc", mount_point_x86_64);
                g_debug("SDL2 FIX: mounting [%s] to [%s]", source, pc_path);

                flatpak_bwrap_add_args (bwrap,
                                        "--ro-bind",
                                        source,
                                        pc_path,
                                        NULL);
                g_free (source);
                g_free (pc_path);
            }
            g_free (full_64_runtime_fix_path);
        }
        else
            g_debug ("SDL2 FIX: %s not supported", full_64_runtime_fix_path);
    }
}


/* wrapper to list/add runtime fixes we'll apply*/
static void
flatpak_run_add_runtime_fixes (FlatpakBwrap *bwrap, gchar *runtime_ref, gchar *flatpak_runtime_install_path, gchar *mount_point_x86_64, gchar *mount_point_i386)
{
    flatpak_run_add_alsa_support_args(
        bwrap,
        runtime_ref,
        flatpak_runtime_install_path,
        mount_point_x86_64,
        mount_point_i386);

    flatpak_run_add_sdl2_support_args(
        bwrap,
        runtime_ref,
        flatpak_runtime_install_path,
        mount_point_x86_64,
        mount_point_i386);

    /* TODO: this is the place to list more fixes */
}


/* Key function to handle fixes we apply in runtimes.
 * Currently we provide re-compiled version of some libraries, in 64-bit and 32-bit format
 * But in general, we can add/remove files or do any other operation we need.
 *
 * These are 3 cases we need to cover:
 *
 *   1. Production case, where APP_ID is appliation ID, and we invoke flatpak via hutchr
 *   2. Debug case when we use --devel flag, and APP_ID is application ID,
 *      which forces Sdk runtime to load. This is important for app troubleshooting,
 *      when we need strace, gdb to debug troublesome application
     3. Debug case when we use APP_ID=RUNTIME/VERS, where we explicitly start shell
 *      in a given runtime, for test purposes
 *
 *  Simplified procedure to cover these cases:
 *
 *  1. Check runtime, if it has 32-bit extensions, and get 32-bit libs mount point
 *     by looking into runtime metafile
 *  2. Determine runtime 64-bit mount point
 *  3. In case we have started application, check if it has additional 32-bit runtime
 *     and apply additional fixes, for 32-bit libs only
 *
 */
static void
flatpak_run_add_runtime_fixes_args (FlatpakBwrap *bwrap) {

    GList *extensions, *l;              /* we use lists to iterate anc check extensions */
    g_auto(GStrv) app_parts = NULL;     /* array to split app ref string, see below */
    g_auto(GStrv) runtime_parts = NULL; /* the same, just for runtime */
    gboolean is_app = FALSE;            /* flag to check if we start app, or just runtime */
    gboolean is_sdk = FALSE;            /* flag to check if runtime is Sdk */


    /* these paths will be passed to the function that adds fixes to bwrap */
    gchar *mount_point_x86_64 = NULL;
    gchar *mount_point_i386 = NULL;

    /* sanity check if flatpak_run_app has filled global vars */
     if (rf_app_ref && rf_runtime_ref)
    {
        /* get application ref "type/appid/arch/ver" into parts:
         *   app_parts[0]  is app type, may be { app, runtime }
         *   app_parts[1]  flatpak app/runtime id
         *   app_parts[2]  app/runtime architecture, may be { i386, x86_64 }
         *   app_parts[3]  app/runtime version
         *
         * NOTE: we use is_app later on to cover the case when app and runtime requires
         *       different 32-bit extensions. Steam with --devel flag is a good example
         */
        g_debug ("RUNTIME FIXES: rf_app_ref=%s", rf_app_ref);
        app_parts = g_strsplit (rf_app_ref, "/", 0);
        is_app = strcmp (app_parts[0], "app") == 0;

        /* runtime_parts, explanation is the same as above, but with the difference that
         * it refers to runtime only, while app_parts can refer to app or runtime, depending on how we call flatpak*/
        g_debug ("RUNTIME FIXES: rf_runtime_ref=%s", rf_runtime_ref);
        runtime_parts = g_strsplit (rf_runtime_ref, "/", 0);
        is_sdk = (g_strrstr (runtime_parts[1], "Sdk") ? TRUE : FALSE);

        /* process runtime first, we have to mount x86_64 fixes for sure, and possibly
         * i386 fixes, if given runtime requires Compat.i386 extension:
         *
         *   1. check if runtime uses Compat.i386 extensions and find mount_point_i386
         *   2. find mount_point_x86_64
         *   3. apply runtime fixes for a given runtime
         */
        if (rf_runtime_metakey)
        {
            extensions = flatpak_list_extensions (rf_runtime_metakey,
                                            runtime_parts[2],
                                            runtime_parts[3]);

            for (l = extensions; l != NULL; l = l->next)
            {
                FlatpakExtension *ext = l->data;
                if (g_strrstr (ext->installed_id, "Compat.i386"))
                {
                    mount_point_i386 = g_strdup_printf ("%s/%s", is_sdk ? "/usr" : "/app", ext->directory);

                    g_debug ("RUNTIME FIXES: found runtime 32-bit extension [%s] mount_point_i386=%s",
                        ext->installed_id,
                        mount_point_i386);
                    break;
                }
            }
            g_list_free (extensions);
        }

        /* To mount 64-bit runtime fixes, we need to check where given runtime keep libraries.
         * NOTE: Older runtimes uses usr/lib, newer one's uses /usr/lib/x86_64-linux-gnu
         * TODO: There must be flatpak system way to check this, not like this
         */

        /* this path is used to verify if installation path has libs and files
         * we provide as fixes. In case runtime don't have them, we won't
         * mount them, bwrap won't be able to create new file in ro mounted
         * runtimes */
        gchar *flatpak_runtime_install_path = g_strdup_printf ("/opt/flatpak/var/lib/flatpak/%s/active/files", rf_runtime_ref);

        /* we use this one to check for older runtimes, that does not provide
         * x86_64-linux-gnu folder for 64-bit libs */
        gchar *flatpak_runtime_install_path_lib = g_strdup_printf ("%s/%s", flatpak_runtime_install_path, "lib/x86_64-linux-gnu");

        if (check_folder_existance (flatpak_runtime_install_path_lib, "RUNTIME FIXES") == 0)
            mount_point_x86_64 = g_strdup ("/usr/lib/x86_64-linux-gnu");
        else
            mount_point_x86_64 = g_strdup ("/usr/lib");

        g_debug ("RUNTIME FIXES: mount_point_x86_64=%s", mount_point_x86_64);

        /* Now we know mount paths for both 32-bit (if any) and 64-bit runtime fixes
         * NOTE: 32-bit fixes will be mounted only if mount_point_i386 != NULL
         */
        flatpak_run_add_runtime_fixes (
            bwrap,
            rf_runtime_ref,
            flatpak_runtime_install_path,
            mount_point_x86_64,
            mount_point_i386);

        g_free(mount_point_x86_64);
        g_free(mount_point_i386);
        g_free(flatpak_runtime_install_path);

        /* In production, we start applications, and applications may require 32-bit
         * extensions too, which may be different than runtime 32-bit extensions.
         *
         * So if we are starting an app, we have to check app extensions too:
         */
        if (is_app)
        {
            if (rf_app_metakey)
            {
                extensions = flatpak_list_extensions (rf_app_metakey,
                                                app_parts[2],
                                                app_parts[3]);

                for (l = extensions; l != NULL; l = l->next)
                {
                    FlatpakExtension *ext = l->data;
                    if (g_strrstr (ext->installed_id, "Compat.i386"))
                    {
                        mount_point_i386 = g_strdup_printf ("%s/%s", is_app ? "/app" : "/usr", ext->directory);

                        g_debug ("RUNTIME FIXES: found app 32-bit extension [%s] fixes mount_point_i386=%s",
                            ext->installed_id,
                            mount_point_i386);

                        /* this path is used to verify if installation path has libs
                         * we provide as fixes. In case runtime don't have them, we won't
                         * mount them, bwrap won't be able to create new file in ro mounted
                         * runtimes */
                        gchar *flatpak_runtime_install_path = g_strdup_printf ("/opt/flatpak/var/lib/flatpak/%s/active/files", ext->ref);

                        /* this applies only on 32-bit fixes, if we have to mount them
                         * on other locations than runtime 32-bit fixes
                         *
                         * NOTE: we still need to pass rf_runtime_ref, because we fix
                         *       runtimes not apps. App just may reuquire extension to
                         *       be added to runtime.
                         */
                        flatpak_run_add_runtime_fixes (
                            bwrap,
                            rf_runtime_ref,
                            flatpak_runtime_install_path,
                            NULL,
                            mount_point_i386);
                        g_free(mount_point_i386);
                    }
                }
                g_list_free (extensions);
            }
        }
    }
}

/* Key function to handle fixes we apply in rapplications.
 * Applications may provide their own set of resources, used only by themselves.
 * In most cases we will have to fix libraries, for example libsecret for Skype
 * 
 * Applications provide these libraries in their flatpak instsallation paths, under
 * "files" folder. These resources get mounted in sandbox in /app folder. 
 * 
 * Structure of "application fixes" is simple, we have folders named by flatpak appID's
 * and inside we provide fixed files, in the same path as in original flatpak installation
 * folder.
 * 
 * Mounting application fixes is straightforward:
 * 
 *   1. check if "application fixes" exist
 *   2. check if it has folder names by appID we are starting
 *   3. recursively list all the folders and files in "appliaction fixes/appId"
 *   4. for each found file, check if such file exists in appID flatpak installation folder
 *   5. then mount all matching files into sandboxed /app folder
 *
 */
static void
flatpak_run_add_application_fixes_args (FlatpakBwrap *bwrap) {

  GList *fixes, *l;
  g_auto(GStrv) app_parts = NULL;     /* array to split app ref string, see below */

  /* sanity check if flatpak_run_app has filled global vars */
  /* check if we are launching application */
  if (rf_app_ref)
  {
    /* breat rf_app_ref into parts, we need flatpak appID, it's app_parts[1] */
    g_debug ("APPLICATION FIXES: rf_app_ref=%s", rf_app_ref);
    app_parts = g_strsplit (rf_app_ref, "/", 0);


    /* check if we provide fixes for this appID */
    gchar *flatpak_appID_fixes_path = g_strdup_printf ("/opt/flatpak/application fixes/%s", app_parts[1]);
    
    if (check_folder_existance(flatpak_appID_fixes_path, "APPLICATION FIXES") == 0)
    {
      /* we provide fixes for this appID
       * for each one of them, check existince in flatpak installation folder
       * and mount in sandbox */

      fixes = NULL;
      find_rec_all_files(flatpak_appID_fixes_path, &fixes);
      if (fixes)
      {
        for (l = fixes; l != NULL; l = l->next)
        {
          char *fix_path_source = (char *)l->data;
          g_debug ("APPLICATION FIXES: FOUND FIX=%s", fix_path_source);

          /* get relative fix path, to flatpak_appID_fixes_path
           * we will use it to check existance of matching file in the flatpak app
           * installation folder, before we call bwrap
           * +1 is to remove leasing / form relative path
           */
          char *fix_relative_path = fix_path_source + strlen(flatpak_appID_fixes_path) + 1;

          /* this is full path to flatpak app installation folder, and file that 
           * needs to be fixed 
           */
          gchar *fix_app_installation_path = g_strdup_printf ("/opt/flatpak/var/lib/flatpak/%s/active/files/%s", rf_app_ref, fix_relative_path);

          /* check if we have mathing file in the flatpak app installation path */
          if (check_file_existance(fix_app_installation_path, "APPLICATION FIXES") == 0)
          {
            gchar *fix_path_dest = g_strdup_printf ("/app/%s", fix_relative_path);
            g_debug ("APPLICATION FIXES: mounting [%s] to [%s]", fix_path_source, fix_path_dest);
            flatpak_bwrap_add_args (bwrap,
                            "--ro-bind",
                            fix_path_source,
                            fix_path_dest,
                            NULL);
            g_free (fix_path_dest);
          }
          g_free (fix_app_installation_path);
        }
        g_list_free (fixes);
      }
      else
        g_debug ("APPLICATION FIXES: no fixes found at=%s", flatpak_appID_fixes_path);

      g_free (flatpak_appID_fixes_path);
    }
  }
}

/* end Axiom patch */
```

### hunk 2

Patch in function `flatpak_run_add_environment_args`

Find this function call to `flatpak_run_add_session_dbus_args`

Insert this code before this function call:

```c
  /* Axiom patch: runtime fixes */
  /* These fixes only makes sense if we provide "runtime fixes" folder
   * otherwise ignore it */
  struct stat sb;
  if (stat("/opt/flatpak/runtime fixes", &sb) == 0 && S_ISDIR(sb.st_mode)) {
    flatpak_run_add_runtime_fixes_args (bwrap);
  }

  /* These fixes only makes sense if we provide "application fixes" folder
   * otherwise ignore it */
  if (stat("/opt/flatpak/application fixes", &sb) == 0 && S_ISDIR(sb.st_mode)) {
    flatpak_run_add_application_fixes_args (bwrap);
  }
  
  /* end Axiom patch */
```

### hunk 3

Patch in function `flatpak_run_setup_base_argv`

Instead of this code:
```c
  group_contents = g_strdup_printf ("%s:x:%d:%s\n"
                                    "nfsnobody:x:65534:\n"
                                    g->gr_name,
                                    gid, g_get_user_name ());
```

Replace with:

```c
  group_contents = g_strdup_printf ("%s:x:%d:%s\n"
                                    "nfsnobody:x:65534:\n"
                                    /* Axiom patch: alsa fixes */
                                    "audio:x:1100:\n",
                                    /* end Axiom patch */
                                    g->gr_name,
                                    gid, g_get_user_name ());
```

IMPORTANT NOTE: flatpak-1.8.2 has slightly different way to fill in group contents, so we have to apply different patch:

Find code:

```c
  group_contents = g_string_new ("");
  g = getgrgid (gid);
  /* if NULL, the primary group is not known outside the container, so
   * it might as well stay unknown inside the container... */
  if (g != NULL)
    g_string_append_printf (group_contents, "%s:x:%d:%s\n",
                            g->gr_name, gid, g_get_user_name ());
  g_string_append (group_contents, "nfsnobody:x:65534:\n");
```

Add additional line below:

```c
  g_string_append (group_contents, "audio:x:1100:\n");
```


### hunk 4

Patch function `flatpak_run_app`

Find this line:

```c
runtime_metakey = flatpak_deploy_get_metadata (runtime_deploy);
```

Insert code below:

```c
  /* Axiom patch: runtime fixes */
  /* share variable we'll need to apply runtime fixes */

    rf_app_ref = g_strdup (app_ref);
    rf_runtime_ref = g_strdup (runtime_ref);
    rf_app_metakey = metakey;
    rf_runtime_metakey = runtime_metakey;

  /* end Axiom patch */
```

## how to create patches out of patched sources

Enter flatpak source folder, create script `create-patches.sh`:

```sh
#!/bin/sh

cat << EOF > xdg-runtime-dir-fix.patch
--- a/common/flatpak-utils.c
+++ b/common/flatpak-utils.c

EOF

diff -u common/flatpak-utils.c.orig common/flatpak-utils.c | tail -n +3 >> xdg-runtime-dir-fix.patch


cat << EOF > mainscreen-progress.patch
--- a/app/flatpak-cli-transaction.c
+++ b/app/flatpak-cli-transaction.c

EOF

diff -u app/flatpak-cli-transaction.c.orig app/flatpak-cli-transaction.c | tail -n +3 >> mainscreen-progress.patch


cat << EOF > runtime-and-application-fixes.patch
--- a/common/flatpak-run.c
+++ b/common/flatpak-run.c

EOF

diff -u common/flatpak-run.c.orig common/flatpak-run.c | tail -n +3 >> runtime-and-application-fixes.patch
```
