#!/bin/sh -ex

# IMPORTANT NOTE: Run this script on an Axiom machine, with flatpak installed

# This script prepares and build flatpak runtime fixes.
# It is necessary to provide such fixes, so flatpak applications can integrate with
# flatpak hutch, which is based on Axiom. 
FIXES="runtime fixes"


# Create folder "runtime fixes"
# In this folder we will place all runtime fixes, we are now going to build
# This is production folder, which will be published in flatpak hutch
rm -rf "runtime fixes"
mkdir  "runtime fixes"


# We need this environment var while instlling flatpak runitimes
# Install procedure will try to get host theme, asking gsettings-desktop-schemas
export XDG_DATA_DIRS=/opt/gsettings-desktop-schemas-3.20.0/share


# List flatpak available SDK runtimes
# Runtimes are available as "Platform" and "Sdk" (TYPE)
# We use "Sdk" runtimes with build essentials (gcc, make, auto tools), to build fixes
# "remote-ls" implies flathub runtimes, we won't use any other
# 

#
# First install all runtimes, this operation may fail due to flathub server error
# By doing installation first, we ensure we can run this script seceral times, and
# continue form the last failed insallation.
# Once runtimes are insalled, building fixes won't fail
# NOTE: we cover i386 runtime and toolcahins
#
IFS=$'\n'
for RUNTIME in $(flatpak remote-ls)
do
  TYPE=$(echo $RUNTIME | awk '{print $1}')
  if [ "$TYPE" == "Sdk" ] || [ "$TYPE" == "i386" ] || [ "$TYPE" == "toolchain-i386" ]; then
    NAME=$(echo $RUNTIME | awk '{print $2}')
    VERS=$(echo $RUNTIME | awk '{print $3}')
    ARCH=$(echo $RUNTIME | awk '{print $4}')

    flatpak install --assumeyes $NAME/$ARCH/$VERS
  fi
done

# NOTE: flatpak remote-ls won't list runtimes which support has expired
#
# Install known expired runtimes, apps may still use it until developers
# finds time to move to newer version
flatpak install --assumeyes org.gnome.Sdk/x86_64/3.30
flatpak install --assumeyes org.gnome.Sdk/x86_64/3.32
flatpak install --assumeyes org.kde.Sdk/x86_64/5.10


# For each installed Sdk runtime:
#   1. create entry in "runtime fixes" folder
#   2. call flatpak to run a script that build fixes, in the given Sdk runtime
#   3. copy built fix inside "runtime fixes"
for RUNTIME in $(flatpak list --columns=ref)
do
  # NOTE: We want to avoid runtimes that has something more than *.Sdk/
  #       like Sdk.Comapat.i386. Unfortunately flatpak list don't have
  #       runtime type, like flatpak remote-ls, so we apply a hack
  if [ ! "${RUNTIME##*Sdk/*}" ]; then
    NAME=$(echo $RUNTIME | cut -d'/' -f1)
    ARCH=$(echo $RUNTIME | cut -d'/' -f2)
    VERS=$(echo $RUNTIME | cut -d'/' -f3)

    mkdir -p "runtime fixes"/$NAME/$VERS

    #
    # Build fixes and fill in "runtime fixes"/$NAME/$VERS

    # 1. setup shared folders for runtime in which we are going to build fixes
    # 
    BUILDDIR=/home/jerry/builddir    # folder to do all the build work
    OUTPUT=$BUILDDIR/fixes           # folder where build results will be stored
    
    mkdir -p $OUTPUT

    # 2. copy and run scripts that builds fixes in the given runtime
    #    NOTE: these scripts shall place build results in  the $OUTPUT folder
    #    NOTE: we search all .sh files, except this script $0
    for FIX in $(ls | grep -v $(basename $0) | grep sh)
    do
        cp "$FIX" $BUILDDIR
        
        # Explanations:
        #   --command="/$BUILDDIR/$FIX" - runs a scrip which builds fix inside Sdk runtime, NOTE: --command don't support additional argument, we need to use envars, to pass runtime name and version to the build script
        #   --filesystem=$BUILDDIR - share host side folder, where we build and collect results
        #   --share=network - share host network, build scripts will download sources
        #   --env=RUNTIME=$NAME-$VERS - pass runtime to script, in general script can build different lib version for each runtime
        #   --env=BUILDDIR=$BUILDDIR - export BUILDDIR, so build script can leverage it
        #   --env=OUTPUT=$OUTPUT - export OUTPUT, so build script can leverage it
        #   --env=BUILDHOST={x86_64, i686} - tell build script to build 64bit or 32bit libs
        
        flatpak run --command="/$BUILDDIR/$FIX" --filesystem=$BUILDDIR --share=network --env=RUNTIME=$NAME-$VERS --env=BUILDDIR=$BUILDDIR --env=OUTPUT=$OUTPUT --env=BUILDHOST=x86_64 $NAME//$VERS

        # few runtimes provides 32bit libraries and toolchains
        # we will build these libraries too, some games (Steam based in particular) still uses 32bit libs
        # NOTE: --allow=multiarch - allow sandbox to run 32bit apps, configure needs to run test programs
        if [ "$NAME-$VERS" == "org.freedesktop.Sdk-19.08" ] ||
           [ "$NAME-$VERS" == "org.freedesktop.Sdk-20.08" ] ||
           [ "$NAME-$VERS" == "org.gnome.Sdk-3.34" ] ||
           [ "$NAME-$VERS" == "org.gnome.Sdk-3.36" ] ||
           [ "$NAME-$VERS" == "org.gnome.Sdk-3.38" ]; then
            flatpak run --command="/$BUILDDIR/$FIX" --allow=multiarch --filesystem=$BUILDDIR --share=network --env=RUNTIME=$NAME-$VERS --env=BUILDDIR=$BUILDDIR --env=OUTPUT=$OUTPUT --env=BUILDHOST=i686 $NAME//$VERS
        fi
    done

    # 3. then we simply copy all results, and clean OUTPUT folder
    cp -r $OUTPUT/* "runtime fixes"/$NAME/$VERS || true
    rm -rf $OUTPUT
  fi
done

# flatpak has patches which behaves this way:
#
#    a. If we just run an application, flatpak will search in "runtime fixes" for a match
#       with .Platform extension
#    b. If we run Sdk runtime, flatpak will search for the .Sdk mathch
#
#    In both cases we provide the same fix, and it's located in the folder with .Sdk suffix
#    To provide the same fix for .Platform runtimes, we just create symlinks.
#
#    For example: org.freedesktop.Platform -> org.freedesktop.Sdk
#
for RUNTIME in $(ls "runtime fixes")
do
    ln -sv $RUNTIME "runtime fixes"/$(basename $RUNTIME .Sdk).Platform
done

# pack production archive
tar czf flatpak_runtime_fixes.tar.gz "runtime fixes"

# NOTE: production folder is: http://mirrors.tassig.com/flatpak/

# Final cleanup
rm -r "runtime fixes"
rm -r $BUILDDIR
