#!/bin/sh -ex

# This script depends on RUNTIME environment variable
# In general, for each runtime we shall know which version of lib to build

if [ "$RUNTIME" = "" ]; then
  echo "define RUNTIME envar in format NAME-VERSION"
  exit
fi

case $RUNTIME in

  org.kde.Sdk-5.10)
    SDL2=SDL2-2.0.7
    VER=7
    ;;

  org.freedesktop.Sdk-1.6 | org.gnome.Sdk-3.26 | org.gnome.Sdk-3.28 | org.kde.Sdk-5.9 | org.kde.Sdk-5.11)
    SDL2=SDL2-2.0.8
    VER=8
    ;;

  org.freedesktop.Sdk-18.08 | org.gnome.Sdk-3.30 | org.gnome.Sdk-3.32 | org.kde.Sdk-5.12)
    SDL2=SDL2-2.0.9
    VER=9
    ;;

  org.freedesktop.Sdk-19.08 | org.gnome.Sdk-3.34 | org.gnome.Sdk-3.36 | org.kde.Sdk-5.13 | org.kde.Sdk-5.14)
    SDL2=SDL2-2.0.10
    VER=10
    ;;

  org.freedesktop.Sdk-20.08 | org.gnome.Sdk-3.38 | org.kde.Sdk-5.15)
    SDL2=SDL2-2.0.12
    VER=12
    ;;
esac

# This program builds sdl2 in a given flatpak SDK runtime, and prepares 
# custom installation which we use to override default sdl2 in production 

# IMPORTANT: Normally this script is invoked by "build all runtime fixes.sh", which exports
#            3 env vars, $BUILDDIR, and $OUTPUT $BUILDHOST. If you intentd to use it outside this scope
#            define these variables manually

# Warn used if these variables are missing
if [ "$BUILDDIR" = "" ] || [ "$OUTPUT" = "" ] || [ "$BUILDHOST" = "" ]; then
    echo "Define BUILDDIR, OUTPUT and BUILDHOST env variables"
    exit
fi

# This is the folder where we collect all build results
PACKAGE=$OUTPUT/$BUILDHOST/$SDL2
mkdir -p $PACKAGE

# Define compiler dependent on BUILDHOST var
case $BUILDHOST in
    x86_64)
        export CC=gcc
        ;;
    i686)
        export PATH=$PATH:/usr/lib/sdk/toolchain-i386/bin/
        export CC=i686-unknown-linux-gnu-gcc
        ;;
    *)
        export CC=gcc
        ;;
esac

# Do all the work in this $BUILDDIR folder
cd $BUILDDIR

# remove any tar.gz files, and download SDL2 sources
curl -O http://mirrors.tassig.com/SDL2/$SDL2.tar.gz

# We are operating under flatpak app SDK runtime
# runtime_info will associate built package with runtime used for building
cat /.flatpak-info | grep runtime= > $PACKAGE/runtime_info

# The whole work is split in 2 stages:

# Stage 1: build and install in the temporary dir
# 
# Form this stage we keep include/ and share/
# IMPORTANT: libSDL2-2.0.so.0.10.0 will have hardcoded --prefix, and won't be usable in runtime
#            sdl2.pc file will also be affected by --prefix, won't be usable
mkdir install/
cd install/
INSTALL_PATH=$(pwd)
cd ../

tar xf $SDL2.tar.gz
cd $SDL2
./configure --prefix="$INSTALL_PATH"
make -j
make install

cd ../
cp -r install/include $PACKAGE/
cp -r install/share $PACKAGE/
rm -r install

# do some cleanup
rm -r $PACKAGE/share/aclocal           # we simply don't need it

# Stage 2: we build it again with default --prefix, we won't do any installation
#          now libSDL2-2.0.so.0.$VER.0 and sdl2.pc will be usable in runtime
cd $SDL2
make clean
./configure 
make -j

cd ../
mkdir -p $PACKAGE/lib/pkgconfig
cp $SDL2/build/.libs/libSDL2-2.0.so.0.$VER.0 $PACKAGE/lib/
cp $SDL2/sdl2.pc $PACKAGE/lib/pkgconfig/

# create symlink in $PACKAGE folder, flatpak sdl2 patch depends on it
ln -sv $SDL2 $OUTPUT/$BUILDHOST/sdl2

# remove unpacked source folder
rm -rf $SDL2


