(*
  in bee strings' heap is static and it is only 1MB long
*)


let log msg = 
  let (sec, nsec) = Posix.clock_gettime_monotonic () in 
  printf "[%Li.%09i] %s\n" sec nsec msg


(* formatting big numbers *)
let human_format n =
   let rec worker ret remain = 
     if ret > 2 || Int64.compare remain 1024L < 0 then (ret, remain)
     else worker (ret + 1) (Int64.div remain 1024L) in
   case worker 0 n
	| (0, number) -> sprintf "%Li" number 
	| (1, number) -> sprintf "%Li Kb" number 
	| (2, number) -> sprintf "%Li Mb" number 
	| (_, number) -> sprintf "%Li Gb" number 
   end

(* measure: string-> (unit-> a)-> int64 *)
let measure description foo = 
	log (sprintf "start task %s" description);
	let (start_secs, start_nano) = Posix.clock_gettime_monotonic () in
	let _ = foo () in
	let (end_secs, end_nano) = Posix.clock_gettime_monotonic () in
	let time_elapsed_seconds = Int64.sub  end_secs  start_secs in 
	log (sprintf "task %s done" description);
	time_elapsed_seconds


let buffer_size = 16384
let random_data = String.build buffer_size

let fill_random_data () =
	let fd = Posix.open "/dev/urandom" [| Posix.O_RDONLY |] in
	let len = Posix.read fd random_data buffer_size 0 in
	log (sprintf "filled buffer with %i bytes of random data\n" len);
	let _ = Posix.close fd in 
	()


(* TODO: no need to try on a trimmed device *)
let trim device () = 
   log "start trimming";
   let _ = Posix.system ("blkdiscard " ^ device) in (* TODO: fails with: not supported *)
   log "done trimming"




let fill_device descr device buffer  = 
	let fd = Posix.open device [| Posix.O_RDWR ; Posix.O_SYNC |] in
	let fill_device_with_data fd buffer () = 
	  let rec worker () = Posix.write fd buffer buffer_size 0 = 0 || worker () in 
	  worker ()
	 in 
	let time_elapsed_seconds = measure descr (fill_device_with_data fd buffer) in 
	let _ = Posix.close fd in
	()



let sequential_reads descr fd chunk_sz device_size =
	let read_device_worker  chunk_sz () =
	  let buffer = String.build chunk_sz in
	  let rec worker () = 
	    let n = Posix.read fd buffer chunk_sz 0 in 
	    if n < 1 then () else worker () in
	  worker () in 
	let time_elapsed_seconds = measure  descr  (read_device_worker  chunk_sz) in 
	let speed = Int64.div  device_size  time_elapsed_seconds in 
	printf "%Li seconds elapsed, total size read: %Li ; speed: %s/s\n" 
      time_elapsed_seconds device_size 
      (human_format speed)


let random_reads descr fd chunk_sz interval device_size =
   let read_device_worker  chunk_sz  () =
     let buffer = String.build chunk_sz in
     let rec worker () = 
       let current_pos = Posix.lseek fd interval Posix.SEEK_CUR in 
       (*printf "current position in the device is %Li\n" current_pos;*)
       if Int64.compare current_pos 0L < 0 then 
         () (* we're done *)
       else
         let _ = Posix.read fd buffer chunk_sz 0 in 
         worker () in
     worker () in 
   let time_elapsed_seconds = measure  descr  (read_device_worker  chunk_sz) in 
   let total_size_read = Int64.mul (Int64.div device_size interval) (Int64.of_int chunk_sz) in 
   let speed = Int64.div  total_size_read  time_elapsed_seconds in 
   printf "%Li seconds elapsed, total size read: %Li ; speed: %s/s\n" 
      time_elapsed_seconds total_size_read 
      (human_format speed)
   
   
   
  

let main = 
   (* TODO: verify first that we can open the device and read and write to it *)
	(*let mdevice = Posix.getenv "SSDPERF_DEVICE" in*)
	let device = "/dev/sdc" in
	let fd = Posix.open device [| Posix.O_RDONLY |] in
	
	let device_size = Posix.lseek fd 0L Posix.SEEK_END in 
	printf "device size is %s\n"  (human_format device_size);
	
	let _ = Posix.lseek fd 0L Posix.SEEK_SET in (* reposition the pointer at the beginning *)
	
	fill_random_data ();
	
	sequential_reads "sequential reads of 4K" fd 4096 device_size;
	let _ = Posix.lseek fd 0L Posix.SEEK_SET in (* reposition the pointer at the beginning *)
	
	random_reads "random reads of 4K with interval 1M" fd 4096 1000000L device_size(*;
	fill_device "fill clean device with initial data" device random_data;
	read_device "sequential read of filled device by 4K chunks" device 4096*)
